unit MemoryGame;

{

Memory 3D - Full Source Code Delphi Game
http://zarko-gajic.iz.hr/memory-3d-full-source-code-delphi-game/

}


interface

uses
  System.SysUtils, System.Classes, Vcl.Controls, System.Generics.Defaults, System.Generics.Collections;

type
  TPlayer = class
  private
    fName: string;
    fClaimedPairs: integer;
    fTries: integer;
  public
    property Name : string read fName write fName;
    property Tries : integer read fTries;
    property ClaimedPairs : integer read fClaimedPairs;
  private
    constructor Create(const name : string);
  end;

  TMField = class
  private
    fValue: integer;
    fHost: TObject;
    fPlayer: TPlayer;
  public
    property Value : integer read fValue;

    property Host : TObject read fHost write fHost;

    property Player : TPlayer read fPlayer;

    constructor Create(const value : integer);
  end;

  TMemoryGameEvent = TNotifyEvent;
  TMemoryGamePlayerEvent = procedure(Sender : TObject; const player : TPlayer) of object;
  TMemoryGameFieldEvent = procedure(Sender : TObject; const mField : TMField) of object;
  TMemoryGameFieldPairEvent = procedure(Sender : TObject; const mField1, mField2 : TMField) of object;

  TMemoryGame = class
  private
    fPlayersCount: integer;
    fPlanesCount: integer;
    fClaimedPairs : integer;
    fPairsCount: integer;
    fCurrentPlane: integer;
    fFieldsCountPerPlane : integer;
    fFirstFieldIndexOnCurrentPlane : integer;
    fGameGridColumns : integer;
    fGameGridRows : integer;
    fFields: TObjectList<TMField>;
    fPlayers: TObjectList<TPlayer>;
    fOpenFirst: boolean;
    fOpenedField: TMField;
    fCurrentPlayer: TPlayer;
    fOnFieldsPaired: TMemoryGameFieldPairEvent;
    fOnFieldClaimed: TMemoryGameFieldEvent;
    fOnFieldOpened: TMemoryGameFieldEvent;
    fOnOpenField: TMemoryGameFieldEvent;
    fOnCloseField: TMemoryGameFieldEvent;
    fOnGameOver: TMemoryGamePlayerEvent;
    fOnNextPlayer: TMemoryGamePlayerEvent;
    fOnGameStart: TMemoryGamePlayerEvent;
    fOnPlayerCreated : TMemoryGamePlayerEvent;
    fOnPlaneChanged : TMemoryGameEvent;
    fOnGameCreated: TMemoryGameEvent;
  private
    property OpenFirst : boolean read fOpenFirst write fOpenFirst;

    property ClaimedPairs : integer read fClaimedPairs;
    function AllPairsClaimed : boolean;

    procedure DoPlaneChanged;
  public
    constructor Create;
    destructor Destroy; override;
  public
    property GameGridColumns : integer read fGameGridColumns;
    property GameGridRows : integer read fGameGridRows;
    property PairsCount : integer read fPairsCount;
    property PlayersCount : integer read fPlayersCount;
    property PlanesCount : integer read fPlanesCount;
    property FieldsCountPerPlane : integer read fFieldsCountPerPlane;

    procedure NewGame(const numberOfPairs, numberOfPlayers, numberOfPlanes, zeroValueShift: integer);

    property FirstFieldIndexOnCurrentPlane : integer read fFirstFieldIndexOnCurrentPlane;
    property CurrentPlane : integer read fCurrentPlane;
    function NextPlane : integer;
    function PrevPlane : integer;

    procedure FieldHostAction(Sender : TObject);

    property Fields : TObjectList<TMField> read fFields;
    property Players : TObjectList<TPlayer> read fPlayers;
    property CurrentPlayer : TPlayer read fCurrentPlayer;

    property OpenedField : TMField read fOpenedField;

    property OnFieldClaimed : TMemoryGameFieldEvent read fOnFieldClaimed write fOnFieldClaimed;
    property OnFieldOpened : TMemoryGameFieldEvent read fOnFieldOpened write fOnFieldOpened;
    property OnOpenField : TMemoryGameFieldEvent read fOnOpenField write fOnOpenField;
    property OnCloseField : TMemoryGameFieldEvent read fOnCloseField write fOnCloseField;
    property OnFieldsPaired : TMemoryGameFieldPairEvent read fOnFieldsPaired write fOnFieldsPaired;
    property OnGameOver : TMemoryGamePlayerEvent read fOnGameOver write fOnGameOver;
    property OnNextPlayer : TMemoryGamePlayerEvent read fOnNextPlayer write fOnNextPlayer;
    property OnPlayerCreated : TMemoryGamePlayerEvent read fOnPlayerCreated write fOnPlayerCreated;
    property OnGameStart : TMemoryGamePlayerEvent read fOnGameStart write fOnGameStart;
    property OnGameCreated : TMemoryGameEvent read fOnGameCreated write fOnGameCreated;
    property OnPlaneChanged : TMemoryGameEvent read fOnPlaneChanged write fOnPlaneChanged;
  end;

implementation

{$region 'TMField' }
constructor TMField.Create(const value: integer);
begin
  fValue := value;
  fPlayer := nil;
end;
{$endregion 'TMField'}

{$region 'TMemoryGame' }
function TMemoryGame.AllPairsClaimed: boolean;
begin
  result := ClaimedPairs = PairsCount;
end;

constructor TMemoryGame.Create;
begin
  OpenFirst := true;

  fFields := TObjectList<TMField>.Create(true);
  fPlayers := TObjectList<TPlayer>.Create(true);
end;

destructor TMemoryGame.Destroy;
begin
  FreeAndNil(fFields);
  FreeAndNil(fPlayers);

  inherited;
end;

procedure TMemoryGame.DoPlaneChanged;
var
  mField : TMField;
begin
  for mField in Fields do mField.Host := nil;

  fFirstFieldIndexOnCurrentPlane := (-1 + CurrentPlane) * FieldsCountPerPlane;

  if Assigned(fOnPlaneChanged) then fOnPlaneChanged(self);
end;

procedure TMemoryGame.NewGame(const numberOfPairs, numberOfPlayers, numberOfPlanes, zeroValueShift: integer);
var
  i, rnd : integer;
  aField : TMField;
  newPlayer : TPlayer;

  procedure CalcGridSize;
  //look for ideal rectangle dimensions (square is ideal) to present fields
  var
    i : integer;
  begin
    fGameGridColumns := 1;
    fGameGridRows := FieldsCountPerPlane;

    if Sqrt(FieldsCountPerPlane) = Trunc(Sqrt(FieldsCountPerPlane)) then
    begin
      fGameGridColumns := Trunc(Sqrt(FieldsCountPerPlane));
      fGameGridRows := Trunc(Sqrt(FieldsCountPerPlane));
      Exit;
    end;

    for i := Trunc(Sqrt(FieldsCountPerPlane)) downto 2 do
      if (FieldsCountPerPlane mod i) = 0 then
      begin
        fGameGridColumns := i;
        fGameGridRows := FieldsCountPerPlane div i;
        Exit;
      end;
  end; //CalcGridSize

begin
  fPairsCount := numberOfPairs; if fPairsCount < 1 then fPairsCount := 1;
  fPlayersCount := numberOfPlayers; if fPlayersCount < 1 then fPlayersCount := 1;
  fPlanesCount := numberOfPlanes; if fPlanesCount < 1 then fPlanesCount := 1;
  fFieldsCountPerPlane := (2 * PairsCount) DIV PlanesCount;

  if ((2 * PairsCount) MOD PlanesCount) <> 0 then
  begin
    raise Exception.Create(Format('Total number of fields "%d" (2 x number of pairs "%d") must be dividable by the number of planes "%d". ' + #13#10#13#10 + 'Game NOT created!', [2 * PairsCount, PairsCount, PlanesCount]));
    Exit;
  end;

  CalcGridSize();

  //players
  Players.Clear;
  for i := 1 to PlayersCount do
  begin
    newPlayer := TPlayer.Create('player ' + i.ToString());
    if Assigned(fOnPlayerCreated) then fOnPlayerCreated(self, newPlayer);
    Players.Add(newPlayer);
  end;
  fCurrentPlayer := Players.First;

  //fields
  Fields.Clear;
  for i := 0 to -1 + PairsCount do
  begin
    // value would be zeroValueShift + 0,0,1,1,2,2,3,4...
    Fields.Add(TMField.Create(zeroValueShift + i));
    Fields.Add(TMField.Create(zeroValueShift + i));
  end;

  //randomize field positions
  Randomize;
  Fields.Sort(TComparer<TMField>.Construct(
    function(const Left, Right : TMField) : integer
    begin
      result := -1 + Random(3);
    end
  ));

  fClaimedPairs := 0;
  fCurrentPlane := 1;
  fOpenedField := nil;
  OpenFirst := true;

  if Assigned(OnGameCreated) then fOnGameCreated(self);
  DoPlaneChanged;
  if Assigned(fOnGameStart) then fOnGameStart(self, CurrentPlayer);
end;

function TMemoryGame.NextPlane: integer;
begin
  Inc(fCurrentPlane);
  if fCurrentPlane > PlanesCount then fCurrentPlane := 1;
  DoPlaneChanged;
end;

function TMemoryGame.PrevPlane: integer;
begin
  Dec(fCurrentPlane);
  if fCurrentPlane < 1 then fCurrentPlane := PlanesCount;
  DoPlaneChanged;
end;

procedure TMemoryGame.FieldHostAction(Sender: TObject);
var
  actionOnField: TMField;
  winner, aPlayer : TPlayer;

  function FieldByHost(const host : TObject) : TMField;
  var
    mf : TMField;
  begin
    result := nil;
    for mf in Fields do
      if mf.Host = host then Exit(mf);
  end;
begin
  actionOnField := FieldByHost(sender);
  if actionOnField = nil then Exit;

  if actionOnField.Player = nil then //not claimed
  begin
    if OpenFirst then
    begin
      fOpenedField := actionOnField;

      if Assigned(fOnOpenField) then fOnOpenField(self, actionOnField);

      OpenFirst := false;
    end
    else //open second
    begin
      if actionOnField = OpenedField then //cannot double open
      begin
        if Assigned(fOnFieldOpened) then fOnFieldOpened(self, actionOnField)
      end
      else
      begin
        Inc(CurrentPlayer.fTries);

        if Assigned(fOnOpenField) then fOnOpenField(self, actionOnField);

        if OpenedField.Value = actionOnField.Value then //we have a match
        begin
          OpenedField.fPlayer := CurrentPlayer;
          actionOnField.fPlayer := CurrentPlayer;

          Inc(fClaimedPairs);
          CurrentPlayer.fClaimedPairs := 1 + CurrentPlayer.ClaimedPairs;

          if Assigned(fOnFieldsPaired) then fOnFieldsPaired(self, OpenedField, actionOnField);

          if AllPairsClaimed then
          begin
            winner := CurrentPlayer; //even if there are other players with the same number of claimed pairs
            for aPlayer in Players do
              if aPlayer.ClaimedPairs > winner.ClaimedPairs then winner := aPlayer;

            if Assigned(fOnGameOver) then fOnGameOver(self, winner);
          end;
        end
        else //no mach pair
        begin
          Sleep(700); //todo: promote interval to property
          if Assigned(OpenedField.Host) AND Assigned(fOnCloseField) then fOnCloseField(self, OpenedField);
          if Assigned(actionOnField.Host) AND Assigned(fOnCloseField) then fOnCloseField(self, actionOnField);

          if CurrentPlayer = Players.Last then
            fCurrentPlayer := Players.First
          else
            fCurrentPlayer := Players[1 + Players.IndexOf(CurrentPlayer)];

          if Assigned(fOnNextPlayer) then fOnNextPlayer(self, CurrentPlayer);
        end;

        fOpenedField := nil;
        OpenFirst := true;
      end;
    end;
  end
  else //already claimed
  begin
    if Assigned(fOnFieldClaimed) then fOnFieldClaimed(self, actionOnField);
  end;
end;

{$endregion}

{$region 'TPlayer' }

constructor TPlayer.Create(const name : string);
begin
  fName := name;
  fTries := 0;
end;

{$endregion}

end.

